| package |
package := Package name: 'MethodWrappers'.
package paxVersion: 0;
	basicComment: 'Andres Otaduy 2002
<aotaduy@frro.utn.edu.ar>

A basic port of the VisualWorks implementation.

Refer to MethodWrapper class comment for details.

Visual Works complete Source Code at: 
http://st-www.cs.uiuc.edu/~brant/Applications/MethodWrappers.html
'.


package classNames
	add: #MethodWrappedObject;
	add: #MethodWrapper;
	add: #MethodWrapperInLiteral;
	add: #MethodWrapperInLiteralTracer;
	yourself.

package methodNames
	add: #CompiledCode -> #isMethodWrapper;
	add: #CompiledMethod -> #unwrappedMethod;
	yourself.

package binaryGlobalNames: (Set new
	yourself).

package globalAliases: (Set new
	yourself).

package allResourceNames: (Set new
	yourself).

package setPrerequisites: (IdentitySet new
	add: '..\..\..\Dolphin Smalltalk 5.0\Object Arts\Dolphin\Base\Dolphin';
	yourself).

package!

"Class Definitions"!

Object subclass: #MethodWrappedObject
	instanceVariableNames: 'inst'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
CompiledMethod variableSubclass: #MethodWrapper
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: 'methods PropertyRegister'!
MethodWrapper variableSubclass: #MethodWrapperInLiteral
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
MethodWrapperInLiteral variableSubclass: #MethodWrapperInLiteralTracer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!

"Global Aliases"!


"Loose Methods"!

!CompiledCode methodsFor!

isMethodWrapper

	^false! !
!CompiledCode categoriesFor: #isMethodWrapper!public! !

!CompiledMethod methodsFor!

unwrappedMethod
	^self! !
!CompiledMethod categoriesFor: #unwrappedMethod!accessing!public! !

"End of package definition"!

"Source Globals"!

"Classes"!

MethodWrappedObject guid: (GUID fromString: '{DC55E223-3FB7-11D7-B0D8-925C86C44478}')!
MethodWrappedObject comment: 'This is an example class
in the class side you can find a method which permits your methodWrappers to "survive" a class recompilation.
you must copy the class method #recompile: of this class to the class you want to persist your method wrappers.
I havent found yet an easy way (without changing the kernel classes) to make a MW survive the modification of the source of a method. This could be easily done on the browser (asuming that you just change your method with this tool ;-)  ). 

For example  evaluate:
(MethodWrapper on:#hello inClass: self) install.
then add an instVar.
you will notice that the method is still wrapped ( by looking at his icon) 

Andres Otaduy 2002.
'!
!MethodWrappedObject categoriesForClass!Kernel-Objects! !
!MethodWrappedObject methodsFor!

hello
	Transcript
		show: 'Hello !! ';
		cr;
		flush! !
!MethodWrappedObject categoriesFor: #hello!public! !

!MethodWrappedObject class methodsFor!

recompile: selector
	"With this you may prevent your methodWrappers from being lost on an instVars change or superclass change (class recompilation) "

	| oldMethod newMethod |
	oldMethod := self compiledMethodAt: selector.
	newMethod := super recompile: selector.
	^oldMethod isMethodWrapper 
		ifTrue: [(oldMethod class on: selector inClass: self) install]
		ifFalse: [newMethod]! !
!MethodWrappedObject class categoriesFor: #recompile:!public! !

MethodWrapper guid: (GUID fromString: '{66ABFB00-BE01-11D6-B0D7-000021FB21B2}')!
MethodWrapper comment: 'This is a port from VisualWorks 5i.
we mantain the clientMethod reference in a property because the Dolphin VM  does not handle very well the addition of instance variables on the CompiledCode Family.
See the class MethodWrapperInLiteral for an efficient implementation.
Andres Otaduy 2002
<aotaduy@frro.utn.edu.ar>

Limitations on this implementation:
	If you recompile a class in some way ( for example adding or removing an instance variable or by changing his superclass) you will  lose your wrappers.
	
Original comment: 

MethodWrappers can add "hidden" behavior to the ordinary method. They can change what is executed before or after the method. For example, a CountMethodWrapper would increment a counter when the method is called, and then forward the call onto the original method. 

While this class is not abstract (i.e., there are no subclassResponsibilities), these method wrappers don''t do anything if you install them. Look at the subclasses for example uses.

To create a new type of method wrapper, you must create a subclass of MethodWrapper and redefine the #beforeMethod and/or #afterMethod as necessary. If you only need to redefine the #beforeMethod, you can redefine the #valueFrom:arguments: method instead (see the CountMethodWrapper). This will give you some added performance since it doesn''t need to create two blocks and send the #valueNowOrOnUnwindDo: message.

To use a method wrapper, send the class a #on:inClass message. This will return a new method wrapper (or nil if this method shouldn''t be wrapped). You can then install the method wrapper by sending the #install message and uninstall by sending the #uninstall message.

WARNING: incorrectly using the wrappers can crash your image, so you should save your image before using them. Most problems occur as a result of installing them on methods that are called either in the #beforeMethod or #afterMethod. (Note: the #value method in Association (VariableBinding in VW2.5) is automatically called by the VM when you refer to a global, class, or pool variable so you should probably never wrap it. I advise against wrapping primitives also.)

Instance Variables:

clientMethod	<CompiledMethod>	the method that is being wrapped
selector	<Symbol>	the selector that the method is listed under in the method dictionary

Class Instance Variables:

methods	<OrderedCollection of: MethodWrapper>	prototype wrappers that are copied when installing new wrappers'!
!MethodWrapper categoriesForClass!Kernel-Methods! !
!MethodWrapper methodsFor!

afterMethod

!

arrayPosition
	self argumentCount == 0 ifTrue: [^0].
	^(1 to: self size)
		detect: [:i | (self realAt: i) == #Array or: [(self realAt: i) == Array]]
		ifNone: [self error: 'Cannot find position for Array']!

at: anIndex
	"Lie about what's in our literal frame. This has the bad side-effect that we 
	can't view the correct information in inspectors, but the good news is that 
	we stop infinite recursion. Good thing the VM doesn't use the basicAt: 
	method :)."

	| object |
	object := super at: anIndex.
	^object == self ifTrue: [self clientMethod] ifFalse: [object]!

basicAt: anIndex
	"Lie about what's in our literal frame. This has the bad side-effect that we 
	can't view the correct information in inspectors, but the good news is that 
	we stop infinite recursion. Good thing the VM doesn't use the basicAt: 
	method :)."

	| object |
	object := super basicAt: anIndex.
	^object == self ifTrue: [self clientMethod] ifFalse: [object]!

beforeMethod!

class: aClass selector: aSelector 
	| position |
	self at: self methodPosition put: self.
	position := self arrayPosition.
	position == 0 ifFalse: [self at: position put: Array].
	methodClass := aClass.
	selector := aSelector!

clientMethod
	^self propertyAt: #clientMethod ifAbsent:[nil]!

clientMethod: anObject
	self propertyAt: #clientMethod put: anObject!

icon
	"Answers an Icon that can be used to represent this object."

	^self class icon!

install
	| definingClass method |
	definingClass := methodClass whichClassIncludesSelector: selector.
	definingClass isNil ifTrue: [^self].
	method := definingClass compiledMethodAt: selector.
	method == self ifTrue: [^self].
	self clientMethod: method.
	sourceDescriptor := self clientMethod sourceDescriptor.
	methodClass addSelector: selector withMethod: self!

isMethodWrapper

	^true!

methodPosition
	^(1 to: self size) detect: 
			[:i | 
			| each |
			each := self realAt: i.
			each == self class methodWrapperSymbol or: [each == self]]
		ifNone: [self error: 'Cannot find position to insert wrapper']!

propertyManager


	^self class propertyManager!

realAt: anIndex
	^super basicAt: anIndex!

resourceType
	^self clientMethod isNil 
		ifTrue: [super resourceType]
		ifFalse: [self clientMethod resourceType]!

uninstall
	methodClass isNil ifTrue: [^self].
	(methodClass compiledMethodAt: selector ifAbsent: [nil]) == self 
		ifTrue: 
			[methodClass == self clientMethod methodClass 
				ifTrue: [methodClass addSelector: selector withMethod: self clientMethod]
				ifFalse: [methodClass removeSelector: selector]]!

uninstallEverything
	| originalMethod |
	methodClass isNil ifTrue: [^self].
	(methodClass compiledMethodAt: selector ifAbsent: [nil]) == self 
		ifTrue: 
			[originalMethod := self unwrappedMethod.
			methodClass== originalMethod methodClass 
				ifTrue: [methodClass addSelector: selector withMethod: originalMethod]
				ifFalse: [methodClass removeSelector: selector]]!

unwrappedMethod
	^self clientMethod unwrappedMethod!

value: anObject withArguments: args
	"Evaluate the receiver with the argument, anObject, as its receiver, and with the argument,
	anArray, as the arguments to the message.

	Primitive failure reasons:
		0 -	anArray is not an Array
		1 -	anArray contains the wrong number of arguments
		2 -	anObject is not a subinstance of the method class of the receiver."

	self beforeMethod.
	^[self clientMethod value: anObject withArguments: args] ensure: [self afterMethod]!

withAllBlockMethodsDo: aBlock
	"Don't recurse into the literal for self, and evaluate for clientMethod so that we can find literals in 
	clientMethods."

	super withAllBlockMethodsDo: aBlock.
	self clientMethod withAllBlockMethodsDo: aBlock! !
!MethodWrapper categoriesFor: #afterMethod!evaluating/before/after!public! !
!MethodWrapper categoriesFor: #arrayPosition!private! !
!MethodWrapper categoriesFor: #at:!private! !
!MethodWrapper categoriesFor: #basicAt:!private! !
!MethodWrapper categoriesFor: #beforeMethod!evaluating/before/after!public! !
!MethodWrapper categoriesFor: #class:selector:!initialize/release!public! !
!MethodWrapper categoriesFor: #clientMethod!accessing!private! !
!MethodWrapper categoriesFor: #clientMethod:!accessing!private! !
!MethodWrapper categoriesFor: #icon!constants!development!public! !
!MethodWrapper categoriesFor: #install!accessing!public! !
!MethodWrapper categoriesFor: #isMethodWrapper!public! !
!MethodWrapper categoriesFor: #methodPosition!private! !
!MethodWrapper categoriesFor: #propertyManager!public! !
!MethodWrapper categoriesFor: #realAt:!private! !
!MethodWrapper categoriesFor: #resourceType!accessing!public! !
!MethodWrapper categoriesFor: #uninstall!accessing!public! !
!MethodWrapper categoriesFor: #uninstallEverything!accessing!public! !
!MethodWrapper categoriesFor: #unwrappedMethod!accessing!public! !
!MethodWrapper categoriesFor: #value:withArguments:!public! !
!MethodWrapper categoriesFor: #withAllBlockMethodsDo:!enumerating!public! !

!MethodWrapper class methodsFor!

canWrap: aSelector inClass: aClass 
	"Test if a method can be wrapped without causing infinite recursion."

	| class method |
	(aClass includesBehavior: MethodWrapper) ifTrue: [^false].
	aClass == BlockClosure 
		ifTrue: 
			[(#(#valueAsUnwindBlockFrom: #ensure:) includes: aSelector) 
				ifTrue: [^false]].
	class := aClass whichClassIncludesSelector: aSelector.
	class isNil ifTrue: [^false].
	method := class compiledMethodAt: aSelector ifAbsent: [nil].
	^method notNil 
		and: [(self primitives includes: method primitiveIndex) not]!

codeStringFor: numArgs 
	| nameString tempsString |
	nameString := 'value'.
	tempsString := numArgs == 0 
				ifTrue: ['t := #()']
				ifFalse: ['(t := #Array new: ' , numArgs printString , ') '].
	1 to: numArgs
		do: 
			[:i | 
			nameString := nameString , 'value: t' , i printString , ' '.
			tempsString := tempsString , (i == 1 ifTrue: [''] ifFalse: [';']) 
						, ' at: ' , i printString 
						, ' put: t' , i printString].
	^nameString , '
	| t |
	' , tempsString , '.
	^' 
		, self methodWrapperSymbol printString 
			, ' value: self withArguments: t'!

createMethodFor: numArgs 
	^self fromCompiledMethod: (Compiler
		compile: (self codeStringFor: numArgs)
			in: self) !

fromCompiledMethod: aMethod
	| newInstance |
	newInstance := self new: aMethod size.
	1 to: aMethod literalCount do: [:i | newInstance literalAt: i put: (aMethod literalAt: i)].
	newInstance
		header: aMethod header;
		methodClass: aMethod methodClass;
		byteCodes: aMethod byteCodes;
		sourceDescriptor: aMethod sourceDescriptor;
		selector: aMethod selector.
	^newInstance!

icon

	^ProtoObject icon!

initialize
	self == MethodWrapper 
		ifTrue: [self allSubclasses do: [:each | each initialize]].
	methods := Dictionary new.
	PropertyRegister := PropertyManager new!

methods
	^methods isNil ifTrue: [methods := LookupTable new] ifFalse: [methods]!

methodWrapperSymbol
	^#'The method wrapper should be inserted in this position'!

nuke
	"MethodWrapper nuke"

	self uninstallAllWrappers.
	self initialize!

on: aSelector inClass: aClass 
	| wrapper |
	(self canWrap: aSelector inClass: aClass) ifFalse: [^nil].
	wrapper := (self methods at: aSelector argumentCount
				ifAbsentPut: [self createMethodFor: aSelector argumentCount]) copy.
	wrapper class: aClass selector: aSelector.
	^wrapper!

primitives
	^#(61	"at:put:"
	71	"new:"
	116	"valueWithReceiver:arguments:"
	81)	"value"!

propertRegister


	^PropertyRegister!

propertyManager


	^PropertyRegister!

uninstallAllWrappers
	"Emergency method to uninstall everything"

	"MethodWrapper uninstallAllWrappers"

	self withAllSubclasses 
		do: [:each | each allInstances do: [:inst | inst uninstallEverything]]! !
!MethodWrapper class categoriesFor: #canWrap:inClass:!public!testing! !
!MethodWrapper class categoriesFor: #codeStringFor:!private! !
!MethodWrapper class categoriesFor: #createMethodFor:!private! !
!MethodWrapper class categoriesFor: #fromCompiledMethod:!public! !
!MethodWrapper class categoriesFor: #icon!public! !
!MethodWrapper class categoriesFor: #initialize!class initialization!public! !
!MethodWrapper class categoriesFor: #methods!accessing!public! !
!MethodWrapper class categoriesFor: #methodWrapperSymbol!private! !
!MethodWrapper class categoriesFor: #nuke!accessing!public! !
!MethodWrapper class categoriesFor: #on:inClass:!instance creation!public! !
!MethodWrapper class categoriesFor: #primitives!private! !
!MethodWrapper class categoriesFor: #propertRegister!public! !
!MethodWrapper class categoriesFor: #propertyManager!public! !
!MethodWrapper class categoriesFor: #uninstallAllWrappers!accessing!public! !

MethodWrapperInLiteral guid: (GUID fromString: '{DC55E221-3FB7-11D7-B0D8-925C86C44478}')!
MethodWrapperInLiteral comment: 'A method wrapper who stores the reference to his client method on the literal frame. 
Suggested by Alejandro Reimondo from Smalltalking http://www.smalltalking.net
See MethodWrapper coment for the reason of this hack.
By Andres Otaduy 2003
<aotaduy@frro.utn.edu.ar>

this hack may be used if you need aditional instanceVariables to save some state.
'!
!MethodWrapperInLiteral categoriesForClass!Kernel-Methods! !
!MethodWrapperInLiteral methodsFor!

clientMethod
	^self realAt: self size!

clientMethod: anObject
	self basicAt: self size put: anObject! !
!MethodWrapperInLiteral categoriesFor: #clientMethod!accessing!private! !
!MethodWrapperInLiteral categoriesFor: #clientMethod:!accessing!private! !

!MethodWrapperInLiteral class methodsFor!

fromCompiledMethod: aMethod
	| newInstance |
	newInstance := self new: aMethod size +1 .
	1 to: aMethod literalCount do: [:i | newInstance literalAt: i put: (aMethod literalAt: i)].
	newInstance
		header: aMethod header;
		methodClass: aMethod methodClass;
		byteCodes: aMethod byteCodes;
		sourceDescriptor: aMethod sourceDescriptor;
		selector: aMethod selector.
	^newInstance! !
!MethodWrapperInLiteral class categoriesFor: #fromCompiledMethod:!public! !

MethodWrapperInLiteralTracer guid: (GUID fromString: '{DC55E222-3FB7-11D7-B0D8-925C86C44478}')!
MethodWrapperInLiteralTracer comment: 'An example of MethodWrapper, it justs prints the class and selector of the wrapped method on the Transcript.
Andres Otaduy 2003
<aotaduy@frro.utn.edu.ar>

	For an example evaluate: 

	(self on: #show inClass: Presenter) install.
	and then open a Workspace (or any window) and look at Transcript.
	to uninstall all wrappers and clean up evaluate:
	MethodWrapper nuke.
'!
!MethodWrapperInLiteralTracer categoriesForClass!Kernel-Methods! !
!MethodWrapperInLiteralTracer methodsFor!

beforeMethod
	Transcript
		display: Time now ; 
		tab; 
		display: self clientMethod ; 
		cr;
		flush! !
!MethodWrapperInLiteralTracer categoriesFor: #beforeMethod!public! !

"Binary Globals"!

"Resources"!

